/*
    Copyright (C) 2003 Fons Adriaensen <fons.adriaensen@skynet.be>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#ifndef __CLXCLIENT_H
#define __CLXCLIENT_H


#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xresource.h>
#include <clthreads.h>


// -------------------------------------------------------------------------------------------


class X_display
{
public:

    X_display (const char *dispname);
    ~X_display (void);

    unsigned long blackpixel (void) const { return BlackPixel (_dpy, _dsn); }
    unsigned long whitepixel (void) const { return WhitePixel (_dpy, _dsn); }

    Display *dpy (void) const { return _dpy; }
    int      dsn (void) const { return _dsn; }
    Colormap dcm (void) const { return _dcm; }
    GC       dgc (void) const { return _dgc; }

    int depth (void) const { return DefaultDepth (_dpy, _dsn); }
    int xsize (void) const { return WidthOfScreen  (ScreenOfDisplay (_dpy, _dsn)); }
    int ysize (void) const { return HeightOfScreen (ScreenOfDisplay (_dpy, _dsn)); }
    int gray  (void) const { return 0; }

    unsigned long  alloc_color (const char *name, unsigned long def);
    XFontStruct   *alloc_font  (const char *name);

private:

    friend class X_rootwin;
    friend class X_window;
    friend class X_handler;

    Display     *_dpy;   
    int          _dsn;
    Colormap     _dcm;
    GC           _dgc;
};


// -------------------------------------------------------------------------------------------


class X_hints
{
public:

    X_hints::X_hints (void) { _sh.flags = 0; _mh.flags = 0; _ch.res_name = 0; _ch.res_class = 0; }

    void position (int x, int y);
    void minsize  (int x, int y);
    void maxsize  (int x, int y);
    void input (int input);
    void state (int state);
    void group (Window group);
    void rname (const char *rname) { _ch.res_name  = (char *) rname; }
    void rclas (const char *rclas) { _ch.res_class = (char *) rclas; }

    friend class X_window;

private:

    XSizeHints  _sh;
    XWMHints    _mh;
    XClassHint  _ch;

};


// -------------------------------------------------------------------------------------------


class X_window
{
public:

    X_window (X_window *pwin, int xpos, int ypos, int xsize, int ysize,
	      unsigned long bgcol, unsigned long bdcol = 0, int bdpix = 0);

    virtual ~X_window (void);    

    X_display*  disp (void) const { return _disp; }

    Display   *dpy (void) const { return _disp->_dpy; } 
    int        dsn (void) const { return _disp->_dsn; }
    Colormap   dcm (void) const { return _disp->_dcm; }
    GC         dgc (void) const { return _disp->_dgc; }
    Window     win (void) const { return _wind; }

    virtual void handle_event (XEvent *E) {}

    int  x_map (void) const { return XMapWindow   (_disp->_dpy, _wind); }
    int  x_mapraised (void) const { return XMapRaised (_disp->_dpy, _wind); }
    int  x_unmap (void) const { return XUnmapWindow (_disp->_dpy, _wind); }
    int  x_clear (void) const { return XClearWindow (_disp->_dpy, _wind); }
    int  x_raise (void) const { return XRaiseWindow (_disp->_dpy, _wind); }
    int  x_move (int xp, int yp) const;
    int  x_resize (int xs, int ys) const;
    int  x_moveresize (int xp, int yp, int xs, int ys) const;
    int  x_set_attrib (unsigned long mask, XSetWindowAttributes *attr) const;
    int  x_set_win_gravity (int gravity) const;
    int  x_set_bit_gravity (int gravity) const;
    int  x_add_events (unsigned long events);
    int  x_rem_events (unsigned long events);
    int  x_set_title (const char *title);
    int  x_set_background (unsigned long color);
    void x_apply (X_hints *hints);

    X_window *pwin (void) { return _pwin; }

    friend class X_rootwin;

private:
  
    X_window (X_display *disp);
    X_window (const X_window&);
    X_window& operator=(const X_window&);
    X_window      *find (Window w);

    unsigned long _ebits;
    X_display     *_disp;
    X_window      *_pwin;
    X_window      *_next;
    X_window      *_list;
    Window         _wind;
};


// -------------------------------------------------------------------------------------------


class X_rootwin : public X_window
{
public:

    X_rootwin (X_display *disp);

    void handle_event (void);
    void handle_event (XEvent *E);
    bool check_done (void) const { return _list == 0; }

private:
};


// -------------------------------------------------------------------------------------------


class X_resman
{
public:

    enum { NDEFOPT = 13 };

    X_resman (void);
    ~X_resman (void);
    X_resman (const X_resman&);
    X_resman& operator=(const X_resman&);

    void X_resman::init (int *argc, char *argv [], char *name,
			 XrmOptionDescRec *opt, int nopt);

    const char *get (const char *res, const char *def) const;
    int   getb (const char *res, int def) const;
    void  geometry (const char *res, int xd, int yd, int bd,
		    int &xp, int &yp, int &xs, int &ys);

    const char *rname (void) { return _rname; }
    const char *rclas (void) { return _rclas; }

private:
  
    char         _rname [64];
    char         _rclas [64];
    char        *_home;  
    XrmDatabase  _xrmdata;

    static XrmOptionDescRec X_resman::defopt [X_resman::NDEFOPT];
};



// -------------------------------------------------------------------------------------------


extern "C" int x_handler_error (Display *dpy, XErrorEvent *err);


class X_handler : public H_thread
{
public:

    X_handler (X_display *disp, Edest *dest, int ipid);
    ~X_handler (void);

    void thr_main (void);
    void next_event (void) { put_event (ITC_ip1q::N_MQ, 1); }

private:

    int  _conn;
};


// -------------------------------------------------------------------------------------------


class X_callback 
{
public:

    enum
    {
	BUTTON_DN   = 1,
	BUTTON_UP   = 2,

	TEXTIP_KEY  = 10,
	TEXTIP_BUT  = 11,
	TEXTIP_MOD  = 12,

	ENUMIP_KEY  = 20,
	ENUMIP_BUT  = 21,
	ENUMIP_MOD  = 22,

	ROTARY_SET  = 30,
	ROTARY_MOVE = 31,
	ROTARY_STOP = 32,

	SLIDER_SET  = 40,
	SLIDER_MOVE = 41,
	SLIDER_STOP = 42,

	BROWSE_SEL  = 50,

	MENUWIN_OPEN = 60,
	MENUWIN_SEL  = 61,
	MENUWIN_CAN  = 62,

	USER_CB      = 1000
    };

    virtual void handle_callb (int /*type*/, X_window */*W*/, XEvent */*E*/) = 0;
};


// -------------------------------------------------------------------------------------------


class X_draw
{
public:

    enum halign { HLEFT = 0, HCENTER = 1, HRIGHT = 2 };
    enum valign { VTOP = 0, VCENTER = 1, VBASE = 2, VBOTTOM = 3 };
 
    X_draw (Display *dpy, Drawable drw, GC gct) :
	_dpy (dpy), _drw (drw), _gct (gct), _xx (0), _yy (0), _hsa (HLEFT), _vsa (VBASE) {}

    ~X_draw (void) {}

    void clearwin (void) { XClearWindow (_dpy, _drw); }
    void setcol (unsigned long c) { XSetForeground (_dpy, _gct, c); }
    void setfunc (int f) { XSetFunction (_dpy, _gct, f); }
    void setline (int d) { XSetLineAttributes (_dpy, _gct, d, LineSolid, CapButt, JoinBevel); } 
  
    void sethsal (int a) { _hsa = a; }
    void setvsal (int a) { _vsa = a; }

    void move (int x, int y) { _xx = x; _yy = y; }
    void rmove (int x, int y) { _xx += x; _yy += y; }
    void draw (int x, int y) { XDrawLine (_dpy, _drw, _gct, _xx, _yy, x, y); _xx = x; _yy = y; }
    void rdraw (int x, int y) { XDrawLine (_dpy, _drw, _gct, _xx, _yy, _xx + x, _yy + y); _xx += x; _yy += y; }
    void drawline (int x0, int y0, int x1, int y1) { XDrawLine (_dpy, _drw, _gct, x0, y0, x1, y1); }

    void drawrect (int x0, int y0, int x1, int y1) { XDrawRectangle (_dpy, _drw, _gct, x0, y0, x1 - x0, y1 - y0); }
    void fillrect (int x0, int y0, int x1, int y1) { XFillRectangle (_dpy, _drw, _gct, x0, y0, x1 - x0, y1 - y0); }

    void drawlines (size_t n, int *x, int *y);
    void drawlines (size_t n, float *x, float *y);

    void setfont (XFontStruct *font) { _font = font; XSetFont (_dpy, _gct, _font->fid); }
    int  textwidth (const char *str);
    int  textheight (void) { return _font->ascent + _font->descent; }
    void drawstring (const char *str);

private:
  
    Display      *_dpy;
    Drawable      _drw;
    GC            _gct;
    XFontStruct  *_font;
    int           _xx;
    int           _yy;
    int           _hsa;
    int           _vsa;
};


// -------------------------------------------------------------------------------------------


class X_linked 
{
private:

    X_linked (X_linked *back, X_linked *forw) :
	_back (back),
	_forw (forw)
	{
	    if (back) back->_forw = this;
	    if (forw) forw->_back = this;
	}

    ~X_linked (void)
	{
	    if (_back) _back->_forw = _forw;
	    if (_forw) _forw->_back = _back;
	}

    virtual void select (void) = 0;

    friend class X_textip;
    friend class X_enumip;

    X_linked *_back;
    X_linked *_forw;
};


// -------------------------------------------------------------------------------------------


class X_button_style
{
public:

    enum { PLAIN = 0, BORDER = 1, RAISED = 2, LED = 0x100 };

    XFontStruct   *font;
    struct 
    {
	unsigned long col0;
	unsigned long col1;
	unsigned long text;
	struct 
	{ 
	    unsigned long parc;
	    unsigned long lite;
	    unsigned long dark;
	} shadow;
    } color;
    struct 
    {
	int x;
	int y;
    } size;
    int type;
};


class X_button : public X_window
{
public:

    enum { MAXLEN = 24 };

    X_button (X_window       *parent,
	      X_callback     *callb,
	      X_button_style *style,
	      int xp,
	      int yp,
	      const char *txt1 = 0,
	      const char *txt2 = 0,
	      int cbid = 0);

    ~X_button (void);

    void set_color (unsigned long col0, unsigned long col1) { _col0 = col0, _col1 = col1; }
    void set_text (const char *txt1, const char *txt2);
    void set_stat (bool stat);
    void redraw (void);
    int  stat (void) const { return _stat; } 
    int  cbid (void) const { return _cbid; }

private:

    void handle_event (XEvent *E);
    void expose (XExposeEvent *E);
    void bpress (XButtonEvent *E);
    void brelease (XButtonEvent *E);

    const X_button_style *_style;
    X_callback     *_callb;
    int             _xs;
    int             _ys;
    int             _cbid;  
    int             _down;
    int             _stat;
    int             _len1;
    int             _len2;
    char            _txt1 [MAXLEN];
    char            _txt2 [MAXLEN];
    unsigned long   _col0;
    unsigned long   _col1;
    unsigned long   _text;
};


// -------------------------------------------------------------------------------------------


class X_rotary_style
{
public:

    enum { NONE, CIRC, LINE };
    enum { CONT, R270, A270 };

    struct
    {
	unsigned long bgnd;
	unsigned long top;
	unsigned long thin;
	unsigned long mark;
	struct
	{
	    unsigned long dark;
	    unsigned long med1;
	    unsigned long lite;
	} shadow;
    } color;

    int mode;
    int mark;
};



class X_rotary : public X_window
{
public:

    X_rotary (X_window *parent,
	      X_callback     *callb,
	      X_rotary_style *style, 
	      int xp, int yp,
	      int rad, int rim,
              int cbid = 0);
 
    ~X_rotary (void);

    void  set_abs (float a);
    float get_abs (void) const { return _am; }
    float get_rel (void) const { return _da; }
    int   cbid (void) const { return _cbid; }

private:

    void handle_event (XEvent *xe);
    void expose (XExposeEvent *e);
    void motion (XMotionEvent *e);
    void bpress (XButtonEvent *e);
    void brelease (XButtonEvent *e);
    void xormark (void);

    X_rotary_style  *_style;
    X_callback      *_callb;
    int      _cbid;
    int      _rad;
    int      _rim;
    int      _xc;
    int      _yc;
    int      _xp;
    int      _yp;
    float    _ap;
    float    _am;
    float    _da;
};


// -------------------------------------------------------------------------------------------


class X_textln_style
{

public:

    XFontStruct   *font;
    struct 
    {
	struct
	{
	    unsigned long bgnd;
	    unsigned long text;
	} normal;
	struct 
	{ 
	    unsigned long bgnd;
	    unsigned long text;
	    unsigned long caret; 
	} focus;
	struct 
	{ 
	    unsigned long lite;
	    unsigned long dark;
	} shadow;
    } color;
};


// -------------------------------------------------------------------------------------------


class X_enip_item
{
public:
    int         _code;
    const char *_text;
};


class X_enumip : public X_window, public X_linked
{
public:

    X_enumip (X_window *parent, 
	      X_callback      *callb,
	      X_textln_style  *style,
	      int xp, int yp,
	      int xs, int ys,
	      X_enip_item     *list,
	      X_linked    *back = 0,
	      X_linked    *forw = 0);

    ~X_enumip (void);

    void set_ind (int ind);
    int  base (void) const { return _y0; };
    int  key  (void) const { return _key; }
    int  ind  (void) const { return _ind; }

    virtual void handle_event (XEvent *xe);

private:

    void spkey (int k);
    void cbkey (int k);

    void expose (XExposeEvent *e);
    void bpress (XButtonEvent *e);
    void keypress (XKeyEvent *e);
    void setfocus (XFocusChangeEvent *e);
    void remfocus (XFocusChangeEvent *e);
    void redraw (void);

    virtual void select (void);

    X_callback        *_callb;
    X_textln_style    *_style;
    X_enip_item       *_list;
    char         *_txt;
    int           _ind;
    int           _max;
    int           _len;
    int           _foc;
    int           _key;
    int           _xs;
    int           _ys;
    int           _x0;
    int           _y0;
};


// -------------------------------------------------------------------------------------------


class X_textip : public X_window, public X_linked
{
public:

    X_textip (X_window         *parent, 
	      X_callback       *callb,
	      X_textln_style   *style,
	      int xp, int yp,
	      int xs, int ys,
	      int len,
	      X_linked    *back = 0,
	      X_linked    *forw = 0);

    ~X_textip (void);

    void set_text (const char *txt);
    void set_color (unsigned long bg, unsigned long fg);
    int  base (void) const { return _y0; };
    int  key (void) const { return _key; }
    const char *text (void) { _txt [_len] = 0; return _txt; }
    void insert (int k);
    void hidden (int k);
    void clear_modified (void) { _flags |= CLEAR; }
    void callb_modified (void) { if (_callb) _flags |= CALLB; }
    X_linked *forw (void) const { return _forw; }
    X_linked *back (void) const { return _back; } 

    virtual void handle_event (XEvent *xe);

private:

    enum { FOCUS = 1, CLEAR = 2, CALLB = 4 };

    void go_sol (void);
    void go_lt (void);
    void del_rt (void);
    void go_eol (void);
    void go_rt (void);
    void del_eol (void);
    void del_all (void);
    void del_lt (void);
    void spkey (int k);
    void cbkey (int k);

    void expose (XExposeEvent *e);
    void bpress (XButtonEvent *e);
    void keypress (XKeyEvent *e);
    void paste (XSelectionEvent *e);
    void setfocus (XFocusChangeEvent *e);
    void remfocus (XFocusChangeEvent *e);
    void redraw (void);
    void xorcursor (void);
    int  findindex (int x);
    int  textwidth (int k);

    virtual void select (void);

    X_callback      *_callb;
    X_textln_style  *_style;
    unsigned long _bg;
    unsigned long _fg;
    int           _flags;
    int           _xs;
    int           _ys;
    char         *_txt;
    char         *_hid;
    int           _max;
    int           _len;
    int           _key;
    int           _x0;
    int           _y0;
    int           _ic;
    int           _xc;
};


// -------------------------------------------------------------------------------------------


class X_menuwin_style
{
public:

    enum { PLAIN = 0, BORDER = 1, RAISED = 2, PDOWN = 256 };

    XFontStruct   *font;
    int            step;
    int            type;
    struct 
    {
	unsigned long bgnd;
	unsigned long text;
	unsigned long mask;
	struct 
	{ 
	    unsigned long parw;
	    unsigned long lite;
	    unsigned long dark;
	} shadow;
    } color;
};


class X_menuwin_item
{
public:

    enum { TITLE = 1, SPACE = 2, LAST = 4, MARK = 8, MASKED = 16 };

    const char      *_text;
    int              _bits;
    X_menuwin_item  *_link;  
};



class X_menuwin : public X_window
{
public:

    enum { MAXITEM = 32, DL = 10, DR = 6 };

    X_menuwin (X_window         *parent,
	       X_callback       *callb,
	       X_menuwin_style  *style,
	       int xp,
	       int yp,
	       X_menuwin_item   *items); 

    ~X_menuwin (void);

    void show (void);
    void hide (void);
    void set_mask (unsigned m);
    void clr_mask (unsigned m);
    int xs (void) { return _xs; }
    int ys (void) { return _ys; }
    int selection (void) { return _isel; }


private:

    void handle_event (XEvent *E);
    void expose (XExposeEvent *E);
    void enter (XEnterWindowEvent *E);
    void leave (XLeaveWindowEvent *E);
    void motion (XPointerMovedEvent *E);
    void bpress (XButtonEvent *E);
    void brelse (XButtonEvent *E);
    void redraw (bool clear);
    void select (int k);
    int  findit (int x, int y);

    const X_menuwin_style *_style;
    X_callback            *_callb;
    X_menuwin_item        *_items;
    int             _xs;
    int             _ys;
    int             _nit;
    int             _isel;
    int             _open;
    int             _ypos [MAXITEM];
    int             _tlen [MAXITEM];
};


// -------------------------------------------------------------------------------------------


class X_scale_style
{

public:

    enum { SEGMAX = 20 };

    int             marg;
    int             nseg;
    int             pix  [SEGMAX + 1];
    float           val  [SEGMAX + 1];
    const char     *text [SEGMAX + 1];
    XFontStruct    *font;
    unsigned long   bg;
    unsigned long   fg;

    void   set_tick (int i, int p, float v, const char *t) { pix [i] = p; val [i] = v; text [i] = t; }
    void   set_text (int i, const char *t) { text [i] = t; }
    int    pixlen (void) const { return pix [0] + pix [nseg] + 1; }
    int    length (void) const { return 2 * marg + pixlen (); }
    int    limit (int i);
    int    calcpix (float x);
    float  calcval (int p);
};


class X_hscale : public X_window
{
public:

    X_hscale (X_window       *parent,
	      X_scale_style  *style,
	      int xp,
	      int yp,
	      int h,
	      int d = 0);

private:

    void handle_event (XEvent *xe);
    void expose (XExposeEvent *e);

    X_scale_style  *_style;
    int      _h;
    int      _d;
};


class X_vscale : public X_window
{
public:

    X_vscale (X_window       *parent,
	      X_scale_style  *style,
	      int xp,
	      int yp,
	      int w, 
	      int d = 0);

private:

    void handle_event (XEvent *xe);
    void expose (XExposeEvent *e);

    X_scale_style  *_style;
    int      _w;
    int      _d;
};


// -------------------------------------------------------------------------------------------


class X_meter_style
{
public:

    unsigned long bg;
    unsigned long off [2];
    unsigned long act [2];
    unsigned long mkoff;
    unsigned long mkact;
};


class X_meter : public X_window
{
public:

    X_meter (X_window *parent, X_meter_style *style, X_scale_style *scale, int xp, int yp, int xs, int ys);
 
    void handle_event (XEvent *xe);
    void set_val (float v);
    void set_ref (float v);

    friend class X_hmeter;
    friend class X_vmeter;

private:

    void expose (XExposeEvent *e);
    void plotsect (int i0, int i1, int imin, int imax, int cl, int cr);
    void plotmarks (int a, int b, int c);

    virtual void psect (GC g, int a, int b) = 0;
    virtual void pmark (GC g, int a) = 0;

    X_meter_style  *_style;
    X_scale_style     *_scale;

    int      _x;
    int      _y;
    int      _d;
    int      _i;
    int      _s;
    int      _d0;
    int      _d1;
    int      _d2;
};


class X_hmeter : public X_meter
{
public:

    X_hmeter (X_window *parent, X_meter_style *style, X_scale_style *scale, int xp, int yp, int h);

private:

    virtual void psect (GC g, int a, int b);
    virtual void pmark (GC g, int a);
};


class X_vmeter : public X_meter
{
public:

    X_vmeter (X_window *parent, X_meter_style *style, X_scale_style *scale, int xp, int yp, int w);

private:

    virtual void psect (GC g, int a, int b);
    virtual void pmark (GC g, int a);
};


// -------------------------------------------------------------------------------------------


class X_slider_style
{
public:

  unsigned long bg;
  unsigned long lite;
  unsigned long dark;
  unsigned long knob;
  unsigned long mark;
  int           w;
  int           h;
};


class X_slider : public X_window
{
public:

    X_slider (X_window *parent, X_callback *callb, X_slider_style *style, X_scale_style *scale, 
              int xp, int yp, int xs, int ys, int cbid = 0);
 
    virtual void set_val (float v) = 0;
    virtual void set_color (unsigned long knob, unsigned long mark) = 0;
    float get_val (void) { return _scale->calcval (_i); }
    int   cbid (void) const { return _cbid; }

    friend class X_hslider;
    friend class X_vslider;

private:

    void brelease (XButtonEvent *E);

    X_callback      *_callb;
    X_slider_style  *_style;
    X_scale_style   *_scale;
 
    int      _x;
    int      _y;
    int      _i;
    int      _d;
    int      _cbid;
    unsigned long  _knob;
    unsigned long  _mark; 
};


class X_hslider : public X_slider
{
public:

    X_hslider (X_window *parent, X_callback *callb, X_slider_style *style, X_scale_style *scale,
               int xp, int yp, int h, int cbid);

    virtual void handle_event (XEvent *E);
    virtual void set_color (unsigned long knob, unsigned long mark);
    virtual void set_val (float v);

private:

    bool domove (int i);
    void plines (void);
    void plknob (int i);
    void erase (int i);
};


class X_vslider : public X_slider
{
public:

    X_vslider (X_window *parent, X_callback *callb, X_slider_style *style, X_scale_style *scale,
               int xp, int yp, int w, int cbid);

    virtual void handle_event (XEvent *E);
    virtual void set_color (unsigned long knob, unsigned long mark);
    virtual void set_val (float v);

private:

    bool domove (int i);
    void plines (void);
    void plknob (int i);
    void erase (int i);
};


// -------------------------------------------------------------------------------------------


class X_browse_style
{
public:

    unsigned long   bg;
    unsigned long   col [4];
    XFontStruct    *font;
    int             dy;
};


class X_browse : public X_window
{
public:

    X_browse (X_window        *parent,
	      X_callback      *callb,
	      X_browse_style  *style,
	      int             xp,
	      int             yp,
	      int             xs,
	      int             ys,
	      int             max_item,
	      int             max_char);

    ~X_browse (void);

    void reset (void) { _n_item = _n_char = 0; _sel = 0; }
    int  add_item (const char *txt, int col, int len = 0);  
    void sort (void);
    void show (void);
    void forw (void);
    void back (void);

    const char *ptr (void) { return _ptr [_sel]; }
    int col (void) { return _col [_sel]; }
    int len (void) { return _len [_sel]; } 


private:

    void handle_event (XEvent *xE);
    void expose (XExposeEvent *E);
    void bpress (XButtonEvent *E);
    void update (bool clear);

    X_browse_style  *_style;
    X_callback      *_callb;
    int      _max_item;
    int      _max_char;
    int      _n_item;
    int      _n_char;
    int      _xs;
    int      _ys;
    int      _nrow;
    int      _nclm;
    int      _clm0;
    int      _clm1;
    int      _sel;
    char    *_buff;
    char   **_ptr;
    short   *_len; 
    short   *_col;
    int     *_ind;
    int     *_dxc;
};


// -------------------------------------------------------------------------------------------


#endif

